#ifndef WORKER_H
#define WORKER_H

#include <fstream>
#include <atomic>
#include <QObject>
#include <QThread>
#include <QTextStream>
#include "plotdata.h"

class MyPulse
{
public:
    ulong histNo=0; // history number
    double timeStamp=0; // time stamp, ns
    double height=0; // pulse amplitude, MeV
    int cellNo=0; // cell number
    Vector3D pos; // position, cm
    MyPulse() {}
    MyPulse(std::string record) {
        timeStamp = std::stod(record.substr(0, 24));
        height = std::stod(record.substr(27, 10));
        pos = Vector3D(std::stod(record.substr(39, 6)),
                       std::stod(record.substr(47, 6)),
                       std::stod(record.substr(55, 6)));
        cellNo = std::stoi(record.substr(68, 6));
        histNo = std::stoi(record.substr(77, 10));
    }
    MyPulse(const QString record) {
        timeStamp = record.mid(0, 24).toDouble();
        height = record.mid(27, 10).toDouble();
        pos = Vector3D(record.mid(39, 6).toDouble(),
                       record.mid(47, 6).toDouble(),
                       record.mid(55, 6).toDouble());
        cellNo = record.mid(68, 6).toInt();
        histNo = record.mid(77, 10).toULong();
    }
};

class Event
{
public:
    Event() {}
    uint64_t EventTimecode;
    uint64_t RunEventTimecode;
    uint64_t EventCounter;
    uint16_t AsicID;

    double EventTimecode_ns;
    double RunEventTimecode_ns;

    ushort chargeLG[32];
    ushort chargeHG[32];
    bool   hit[32];
};

class EnabledChannel
{
private:
    // ASIC ID, Ch num
    QMap<QPair<int, int>, QPair<int, int>> CHPairMap;
    QMap<QPair<int, int>, int> CHIDmap;
public:
    EnabledChannel() {}
    EnabledChannel(const QList<Channel> chset)
    {
        for (int i=0;i<chset.size();i++)
        {
            if(chset[i].enabled){
                CHPairMap.insert(qMakePair(chset[i].ASICID[0], chset[i].chNum[0]), qMakePair(chset[i].ASICID[1], chset[i].chNum[1]));
                CHPairMap.insert(qMakePair(chset[i].ASICID[1], chset[i].chNum[1]), qMakePair(chset[i].ASICID[0], chset[i].chNum[0]));
                CHIDmap[qMakePair(chset[i].ASICID[0], chset[i].chNum[0])] = i;
                CHIDmap[qMakePair(chset[i].ASICID[1], chset[i].chNum[1])] = i;
            }
        }
    }

    // getter
    int getID(const QPair<int, int> ch) const
    {
        if (CHIDmap.contains(ch))
            return CHIDmap[ch];
        else
            return -1;
    }
    QPair<int, int> getOpposite(const QPair<int, int> ch) const
    {
        if (CHPairMap.contains(ch))
            return CHPairMap[ch];
        else
            return qMakePair(-1, -1);
    }
};

class Worker : public QThread
{
    Q_OBJECT
private:
    const Setup* config;
    PlotData* plotdata;
    void run() override;
    std::atomic<bool> stopped;
    std::atomic<bool> exitted;

    QFile infile;
    std::ofstream outfile;

    EnabledChannel enabledChs;

    bool processCoincidence(const std::vector<MyPulse>& pulses, Cone& newCone);
    bool processCoincidence(const std::vector<Event>& pulses, Cone& newCone);
//    void readCSV(std::vector<Cone>& cones, QTextStream& in);
    const int eventSize = 152; // 152 bytes
    void readBIN(std::vector<Cone>& cones, QDataStream& in);
    int DecodeCITIROCRowEvents(const char* buf, const int bufSize, Event& DataCITIROCA);
public:
//    explicit Worker(QObject *parent = nullptr);
    Worker(QObject *parent, const Setup* config_, PlotData* plotdata_);

//signals:

public slots:
    void handleStart();
    void handleStop();
    void handlePause();
};

#endif // WORKER_H
